/********************************************************
***
*** JK Microsystems   http://www.jkmicro.com
*** Timer Interrupt Module Example
*** Adam Yergovich 
*** 5-24-2006
*** 
*** An example of using the 16 bit timer 2.  This example
*** is about as fast as you can make the timer without
*** throwing the default kernel into a panic.  It yields
*** roughly a 3.3 kHz timer with a period of about 300uS.
*** It should be noted that under normal circumstances, 
*** you will likely want your timer to run slower.  To 
*** do this, change the load value that goes into the 
*** timer register below.  A load value of 0x4C should be
*** considered a limit.  I should point out as well that 
*** users should take consideration when using the toggle
*** method I have below, as it could vary your output 
*** depending on the load value being odd or even.
***
*** Compiled with the command: 
*** ./gccarm -c -O2 -DMODULE -D__KERNEL__ -isystem /KNOPPIX/kernel/linux-2.4.21/include gpio_int_2.c
*** where "./gccarm" is a link to the gcc cross compiler
*** and the isystem tag points to the kernel files for 
*** the OmniFlash.
*** 
*** Much of this was derived from the wonderful example 
*** by Peter Salzman found here: 
*** http://www.faqs.org/docs/kernel/index.html
***
********************************************************/


/* The necessary header files */
/* Standard in kernel modules */

/* Make certain it is pointing to the right files when you compile!!!!*/

#include "/KNOPPIX/kernel/linux-2.4.21/include/linux/kernel.h"               
#include "/KNOPPIX/kernel/linux-2.4.21/include/linux/module.h"               
#include "/KNOPPIX/kernel/linux-2.4.21/include/linux/tty.h"      
#include "/KNOPPIX/kernel/linux-2.4.21/include/linux/sched.h"

/* We want an interrupt */
#include "/KNOPPIX/kernel/linux-2.4.21/include/linux/interrupt.h"
#include "/KNOPPIX/kernel/linux-2.4.21/include/asm/io.h"




#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

/* Global variables we will need */

static struct file_operations fops;
static int Major;
int i = 0;
int j = 0;
int status = 0;
void *statusAdr, *map_base, *virt_addr_dir, *virt_addr_IO, *virt_addr_clear;


/*
 Print String function from Peter Salzman's tutorial.  Allows for printing to console no matter its location.
*/

void print_string(char *str)
{
   struct tty_struct *my_tty;
   my_tty = current->tty;           // The tty for the current task

   if (my_tty != NULL) { 

      
      (*(my_tty->driver).write)(
         my_tty,                 // The tty itself
         0,                      // We don't take the string from user space
         str,                    // String
         strlen(str));           // Length

     
      (*(my_tty->driver).write)(my_tty, 0, "\015\012", 2);
   }
}



/* Our interrupt handler */

void irq_timer(int irq, void *dev_id, struct pt_regs *regs)
{

  i++;
  //simple toggle for the IO line
  if(i%2)
    *((unsigned long *) virt_addr_IO) = 0x1;    
  else
    *((unsigned long *) virt_addr_IO) = 0x0;

  *((unsigned long *) virt_addr_clear) = 0x1;	//clear the interrupt

}



/* Initialize the module - register the IRQ handler */
int init_module()
{ 
   

    Major = register_chrdev(0, "Text_Out_Device", &fops); 

    if (Major == -1) {
        print_string(" Dynamic Major number allocation failed\n");
        return Major;
    }
    /* Interrupt 5 is the block interrupt for timer 2.
     * SA_SHIRQ means we're willing to have other handlers on this IRQ.
     * SA_INTERRUPT can be used to make the handler into a fast interrupt. 
     */
   status = request_irq(5, 
                         (void *)irq_timer, 
                         SA_INTERRUPT, 
                         "Timer", NULL);
   /* Check for errors */   
   if (status == -EBUSY) {
        print_string("IRQ number allocation failed\n");
        return status;  
    }
   if (status == -ENOMEM){
	print_string("No memory error \n");
	return status;
   }
   if (status == -EINVAL){
	print_string("Invalid error \n");
	return status;
   }
   if (status < 0){
	print_string("Irq is Invalid \n");
	return status;
   }

   
   map_base = __ioremap(0x80840014,32,0);		//map physical address for IO
   *((unsigned long *) map_base) = 0x1; 		//set to port B.0 to output

   map_base = __ioremap(0x80810020,32,0);		//load register address
   *((unsigned long *) map_base) = 0x4C;		//the reload value for the timer.  Greater than 0x4C and less than or equal to 0xFFFF.

   map_base = __ioremap(0x80810028,32,0);
   *((unsigned long *) map_base) = 0xC8;		//Enable 508KHz timer, using the load value above 

   virt_addr_clear = __ioremap(0x8081002C,32,0);   	//Timer 2 interrupt clear address setup
   virt_addr_IO = __ioremap(0x80840004,32,0);		//Port B data register address setup



   /*
   * Probably want to put some additional error checking here.
   * May want to clear any interrupts.
   * Additionally this program does not check if the status of 
   * the port is changed after module insertion.  End users will 
   * likely want to ensure such.  
   */


   print_string("Timer interrupt module inserted, timer started");

   return 0;
}

/* Cleanup */
void cleanup_module()
{
   printk("Removing Module \n");
   free_irq(5, NULL);			   //make sure to free up the IRQ
   unregister_chrdev(Major, "OurDevice");  //make sure to unregister the device
}
